"""FrameNet LexUnit class."""

__all__ = ['LexUnit']


class LexUnit():
    """FrameNet LexUnit class."""

    def __init__(self, frame, _id=None, name=None):
        """Constructor."""
        self.__id = _id
        self._name = name
        self._frame = frame

    @property
    def _id(self):
        """Return the lexunit ID."""
        return self.__id

    @property
    def name(self):
        """Return the lexunit name."""
        return self._name

    @property
    def pos(self):
        """Return the part of speech (POS) of the lexical unit."""
        if self._name is None:
            return None
        return self._name.split('.')[1]

    @property
    def frame(self):
        """Return the frame object corresponding to the lexunit."""
        return self._frame

    @_id.setter
    def _id(self, _id):
        self.__id = _id

    @name.setter
    def name(self, name):
        self._name = name



"""FrameNet AnnotationSet's Target class."""

import logging

__all__ = ['Target']

logger = logging.getLogger(__name__)


def _extract_flat_target_pnw_labels(target_indexes, pnw_labels_by_indexes):
    target_pnw_labels = [labels for indexes, labels
                         in pnw_labels_by_indexes.items()
                         if indexes in target_indexes]
    return [item for sublist in target_pnw_labels for item in sublist]


def _extract_target_pos_tags(target_indexes, pnw_labels_by_indexes):
    flat_target_pnw_labels = _extract_flat_target_pnw_labels(
        target_indexes, pnw_labels_by_indexes)
    return [label for label in flat_target_pnw_labels
            if label.layer.name == 'PENN' or label.layer.name == 'BNC']


def _extract_target_string(text, indexes):
    for (start, end) in indexes:
        if start == -1 or end == -1:
            logger.debug(
                'Target indexes are not specified in sentence: {}'
                .format(text))
            return ''
    return ' '.join([text[start: end+1] for (start, end) in indexes])


def _extract_target_indexes(fn_labels):
    return sorted([(label.start, label.end) for label in fn_labels
                   if label.layer.name == 'Target'])


class Target():
    """FrameNet target class."""

    def __init__(self, string=None, lexunit=None, indexes=None, pos_tags=None):
        """Constructor."""
        self._string = string
        self._lexunit = lexunit
        self._indexes = indexes
        self._pos_tags = pos_tags

    @classmethod
    def from_fn_data(cls, fn_labels, text, pnw_labels_by_indexes, lexunit):
        """Return Target instance generated from FrameNet data."""
        indexes = _extract_target_indexes(fn_labels)
        string = _extract_target_string(text, indexes)
        pos_tags = _extract_target_pos_tags(indexes, pnw_labels_by_indexes)
        return cls(string, lexunit, indexes=indexes, pos_tags=pos_tags)

    @property
    def string(self):
        """Return the string of the target."""
        return self._string

    @property
    def lexunit(self):
        """Return the lexunit corresponding to the target."""
        return self._lexunit

    @property
    def indexes(self):
        """Return a (sorted) list of index tuples (start, end)."""
        return sorted(self._indexes)

    @property
    def pos_tags(self):
        """Return a list of PENN or BNC POS tags for the given target."""
        return self._pos_tags

    @string.setter
    def string(self, string):
        self._string = string

    @lexunit.setter
    def lexunit(self, lexunit):
        self._lexunit = lexunit

    @indexes.setter
    def indexes(self, indexes):
        self._indexes = indexes

    @pos_tags.setter
    def pos_tags(self, pos_tags):
        self._pos_tags = pos_tags



"""Marshalling utils.

A set of functions potentially shared across marshallers.
"""

__all__ = ['marshall_sent_dict', 'get_sent_num', 'get_sent_dict',
           'get_start_index', 'get_end_index']


def marshall_sent_dict(sent_dict, sent_filepath):
    """Marshall a {pyfn.sentence.text:sentence_num} dict to stream."""
    with open(sent_filepath, 'w', encoding='utf-8') as sent_stream:
        for sent_text in sorted(sent_dict.keys()):
            print(sent_text, file=sent_stream)


def get_sent_num(text, sent_dict):
    """Return the sentence number stored in the dict.

    The sent_dict is a {text:sent_num} dictionary. If the sentence text
    is not in the dictionary, the function will add it with a sent_num value
    corresponding to the current length of the sent_dict.
    """
    if text not in sent_dict:
        sent_dict[text] = len(sent_dict)
    return sent_dict[text]


def get_sent_dict(sent_filepath):
    """Return a {text:sent_num} dictionary.

    Given an absolute path to a .sentences file
    """
    sent_dict = {}
    sent_iter = 0
    with open(sent_filepath, 'r', encoding='utf-8') as sent_stream:
        for line in sent_stream:
            line = line.rstrip()
            sent_dict[sent_iter] = line
            sent_iter += 1
    return sent_dict


def get_end_index(token_num, tokens, text):
    """Get the ending char index of a given token_num in text.

    The ending char index corresponds to the last char index of a token
    identified by its token_num, given a text tokenized in a list of tokens.
    """
    return get_start_index(token_num, tokens, text) \
        + len(tokens[token_num]) - 1


def get_start_index(token_num, tokens, text):
    """Get the starting char index of a given token_num in text.

    The starting char index corresponds to the first char index of a token
    identified by its token_num, given a text tokenized in a list of tokens.
    """
    start = 0
    for token_index, token in enumerate(tokens):
        while text[start] != token[0]:
            start += 1
        if token_index == token_num:
            return start
        start += len(token)



"""Unmarshalling BIOS tagging format to pyfn.AnnotationSet objects."""

import pyfn.utils.marshalling as marsh_utils

from pyfn.exceptions.parameter import InvalidParameterError

from pyfn.models.annotationset import AnnotationSet
from pyfn.models.frame import Frame
from pyfn.models.label import Label
from pyfn.models.labelstore import LabelStore
from pyfn.models.layer import Layer
from pyfn.models.lexunit import LexUnit
from pyfn.models.sentence import Sentence
from pyfn.models.target import Target

__all__ = ['unmarshall_annosets']


def _has_fe_labels(lines):
    for line in lines:
        line_split = line.split('\t')
        if line_split[14] != '_' and line_split[14] != 'O':
            return True
    return False


def _get_labelstore(lines, tokens, text, labelstore):
    # Multiple cases to handle:
    # 1. S-tag
    # 2. B-tag
    # 3. continuous I-tag (coming right after the same B or I tag)
    # 4 discontinuous I-tag (not coming after the same B or I tag)
    if not _has_fe_labels(lines):
        return labelstore
    for line in lines:
        line_split = line.split('\t')
        if line_split[14].startswith('S-'):
            label = Label(name=line_split[14][2:], layer=Layer(name='FE'),
                          start=marsh_utils.get_start_index(
                              int(line_split[0])-1, tokens, text),
                          end=marsh_utils.get_end_index(int(line_split[0])-1,
                                                        tokens, text))
            labelstore.labels.append(label)
        elif line_split[14].startswith('B-'):
            label_name = line_split[14][2:]
            label = Label(name=label_name, layer=Layer(name='FE'),
                          start=marsh_utils.get_start_index(
                              int(line_split[0])-1, tokens, text))
            for iline in lines[lines.index(line)+1:]:
                iline_split = iline.split('\t')
                if iline_split[14].startswith('I-') \
                 and iline_split[14][2:] == label_name:
                    label.end = marsh_utils.get_end_index(
                        int(iline_split[0])-1, tokens, text)
                    continue
                else:
                    labelstore.labels.append(label)
                    return _get_labelstore(lines[lines.index(iline):],
                                           tokens, text, labelstore)
            labelstore.labels.append(label)
            return labelstore
        elif line_split[14].startswith('I-'):
            raise InvalidParameterError(
                'Discontinuous FE annotation detected '
                'for line [{}] in sentence \'{}\'. Discontinuous FEs are not '
                'supported by pyfn for the BIOS tagging format'.format(
                    line, ' '.join(tokens)))
    return labelstore


def _update_annoset_target(annoset, tokens, token_index, lexunit,
                           last_target_token_index):
    if not annoset.target:
        annoset.target = Target(
            lexunit=lexunit,
            indexes=[(marsh_utils.get_start_index(token_index-1, tokens,
                                                  annoset.sentence.text),
                      marsh_utils.get_end_index(token_index-1, tokens,
                                                annoset.sentence.text))])
    else:
        if token_index - last_target_token_index > 1:
            annoset.target.indexes.append(
                (marsh_utils.get_start_index(token_index-1, tokens,
                                             annoset.sentence.text),
                 marsh_utils.get_end_index(token_index-1, tokens,
                                           annoset.sentence.text)))
        else:
            annoset.target.indexes = [
                (annoset.target.indexes[
                    len(annoset.target.indexes)-1][0],
                 marsh_utils.get_end_index(token_index-1, tokens,
                                           annoset.sentence.text))]


def _create_annoset(lines, sentence, annoset_id):
    annoset = AnnotationSet(_id=annoset_id, sentence=sentence)
    last_target_token_index = 0
    _tokens = [line.split('\t')[1] for line in lines]
    tokens = sentence.text.split()
    if len(tokens) != len(_tokens):
        raise InvalidParameterError('Number of tokens in sentence do not '
                                    'align between .bios and .sentences file')
    for line in lines:
        line_split = line.split('\t')
        if line_split[12] != '_':
            _update_annoset_target(annoset, tokens, int(line_split[0]),
                                   LexUnit(name=line_split[12],
                                           frame=Frame(name=line_split[13])),
                                   last_target_token_index)
            last_target_token_index = int(line_split[0])
    annoset.labelstore = _get_labelstore(lines, tokens, sentence.text,
                                         LabelStore(labels=[]))
    return annoset


def _create_sent_dict(bios_filepath):
    sent_dict = {}
    with open(bios_filepath, 'r', encoding='utf-8') as bios_stream:
        tokens = []
        sent_index = -1
        for line in bios_stream:
            line = line.strip()
            if line != '':
                tokens.append(line.split('\t')[1])
                sent_index = int(line.split('\t')[6])
            else:
                sent_dict[sent_index] = ' '.join(tokens)
                tokens = []
    return sent_dict


def unmarshall_annosets(bios_filepath, sent_filepath=None):
    """Unmarshall a BIOS-tagged file to pyfn.AnnotationSet objects.

    If the sent_filepath is specified, the sentences dict will be created
    from the .sentences file. Else, it will be created from the tokens in
    the .bios file.
    """
    annosets = []
    if sent_filepath:
        sent_dict = marsh_utils.get_sent_dict(sent_filepath)
    else:
        sent_dict = _create_sent_dict(bios_filepath)
    with open(bios_filepath, 'r', encoding='utf-8') as bios_stream:
        index = -1
        annoset_id = 0
        lines = []
        for line in bios_stream:
            line = line.strip()
            if line != '':
                line_split = line.split('\t')
                sent_index = int(line_split[6])
                if sent_index != index:
                    sentence = Sentence(_id=sent_index,
                                        text=sent_dict[sent_index])
                    index = sent_index
                lines.append(line)
            else:
                annoset = _create_annoset(lines, sentence, annoset_id)
                annoset_id += 1
                annosets.append(annoset)
                lines = []
    return annosets



"""Unmarshall SEMAFOR .frame.elements files."""

import pyfn.utils.marshalling as marsh_utils

from pyfn.models.annotationset import AnnotationSet
from pyfn.models.frame import Frame
from pyfn.models.label import Label
from pyfn.models.labelstore import LabelStore
from pyfn.models.layer import Layer
from pyfn.models.lexunit import LexUnit
from pyfn.models.sentence import Sentence
from pyfn.models.target import Target


__all__ = ['unmarshall_annosets']


def _get_labelstore(line_split, text):
    labels = []
    if len(line_split) <= 8:
        return LabelStore(labels=[])
    for index, item in enumerate(line_split[8:]):
        if index % 2 == 0:
            label = Label(name=item, layer=Layer(name='FE'))
        else:
            start_token = _get_start_token(line_split[8:][index])
            end_token = _get_end_token(line_split[8:][index])
            label.start = marsh_utils.get_start_index(
                start_token, text.split(), text)
            label.end = marsh_utils.get_end_index(
                end_token, text.split(), text)
            labels.append(label)
    return LabelStore(labels=labels)


def _get_target(start_token, end_token, luname, framename, text):
    lexunit = LexUnit(frame=Frame(name=framename), name=luname)
    if end_token - start_token < 2:
        indexes = [(marsh_utils.get_start_index(start_token, text.split(),
                                                text),
                    marsh_utils.get_end_index(end_token, text.split(), text))]
    else:
        indexes = [(marsh_utils.get_start_index(start_token, text.split(),
                                                text),
                    marsh_utils.get_end_index(start_token, text.split(),
                                              text)),
                   (marsh_utils.get_start_index(end_token, text.split(),
                                                text),
                    marsh_utils.get_end_index(end_token, text.split(),
                                              text))]
    return Target(lexunit=lexunit, indexes=indexes)


def _get_end_token(token_indexes):
    if '_' in token_indexes:
        return int(token_indexes.split('_')[1])
    if ':' in token_indexes:
        return int(token_indexes.split(':')[1])
    return int(token_indexes)


def _get_start_token(token_indexes):
    if '_' in token_indexes:
        return int(token_indexes.split('_')[0])
    if ':' in token_indexes:
        return int(token_indexes.split(':')[0])
    return int(token_indexes)


def unmarshall_annosets(semafor_filepath, sent_filepath):
    """Unmarshall a semafor .frame.elements file to pyfn.AnnotationSets."""
    annosets = []
    sent_dict = marsh_utils.get_sent_dict(sent_filepath)
    with open(semafor_filepath, 'r', encoding='utf-8') as semafor_stream:
        for line in semafor_stream:
            line = line.strip()
            line_split = line.split('\t')
            sent_index = int(line_split[7])
            sentence = Sentence(_id=sent_index,
                                text=sent_dict[sent_index])
            target = _get_target(_get_start_token(line_split[5]),
                                 _get_end_token(line_split[5]),
                                 line_split[4],
                                 line_split[3],
                                 sent_dict[sent_index])
            labelstore = _get_labelstore(line_split, sent_dict[sent_index])
            annoset = AnnotationSet(sentence=sentence, target=target,
                                    labelstore=labelstore)
            annosets.append(annoset)
    return annosets



"""Files utils."""

import os

from pyfn.exceptions.parameter import InvalidParameterError

__all__ = ['get_bios_filepath', 'get_sent_filepath', 'get_semafor_filepath',
           'create_parent_dir_if_not_exists', 'get_bios_semeval_filepath',
           'get_ancestors_filepath', 'get_frame_parents_filepath',
           'get_rolemappings_filepath', 'get_fr_relation_xml_filepath']


def get_semafor_filepath(target_dirpath, splits_name):
    """Return the absolute path to the SEMAFOR file.

    If splits_name is 'train', function will return train.frame.elements.
    If splits_name is 'dev' or 'test', function will return train.frames.
    Both files will be seeked under the specified target_dirpath.
    """
    if splits_name not in ['train', 'dev', 'test']:
        raise InvalidParameterError('Unsupported splits_name: {}'.format(
            splits_name))
    if splits_name == 'train':
        return os.path.join(target_dirpath, 'train.frame.elements')
    return os.path.join(target_dirpath, '{}.frames'.format(splits_name))


def get_bios_filepath(target_dirpath, splits_name):
    """Return the absolute path of the {splits_name}.bios file.

    Look for the file under the specified taret_dirpath.
    """
    return os.path.join(target_dirpath, '{}.bios'.format(splits_name))


def get_bios_semeval_filepath(target_dirpath, splits_name):
    """Return the absolute path of the {splits_name}.bios.semeval file.

    Look for the file under the specified taret_dirpath.
    """
    return os.path.join(target_dirpath, '{}.bios.semeval'.format(splits_name))


def get_sent_filepath(target_dirpath, splits_name):
    """Return the absolute path of the {splits_name}.sentences file.

    Look for the file under the specified target_dirpath.
    """
    return os.path.join(target_dirpath, '{}.sentences'.format(splits_name))


def create_parent_dir_if_not_exists(filepath):
    """Create the parent directory of the input filepath if not exists."""
    os.makedirs(os.path.dirname(filepath), exist_ok=True)


def get_ancestors_filepath(target_dirpath):
    """Return the absolute path to the ancestors.csv file."""
    os.makedirs(target_dirpath, exist_ok=True)
    return os.path.join(target_dirpath, 'ancestors.csv')


def get_frame_parents_filepath(target_dirpath):
    """Return the absolute path to the frame_parents.csv file."""
    os.makedirs(target_dirpath, exist_ok=True)
    return os.path.join(target_dirpath, 'frame_parents.csv')


def get_rolemappings_filepath(target_dirpath):
    """Return the absolute path to the frame_parent_rolemappings.csv file."""
    os.makedirs(target_dirpath, exist_ok=True)
    return os.path.join(target_dirpath, 'frame_parent_rolemappings.csv')


def get_fr_relation_xml_filepath(splits_dirpath):
    """Return the absolute path to the frRelation.xml file given splits_dirpath."""
    return os.path.join(splits_dirpath, 'frRelation.xml')



"""Constants used by pyfn."""

import os

__all__ = ['FN_XML_NAMESPACE', 'DATA_DIR', 'SENT_LAYERS', 'ANNO_LAYERS']


FN_XML_NAMESPACE = {'fn': 'http://framenet.icsi.berkeley.edu'}

DATA_DIR = os.path.join(os.path.dirname(__file__), '..', '..', 'data')

ANNO_LAYERS = ['FE', 'PT', 'GF', 'Target']  # layers stored in the
# annoset.labelstore
SENT_LAYERS = ['PENN', 'BNC']  # layers stored in the sentence.labelstore

HIERARCHY_RELATION_TYPES = ['Inheritance', 'SubFrame']



"""FrameNet Corpus object class."""

__all__ = ['Corpus']


class Corpus():
    """FrameNet Corpus class."""

    def __init__(self, _id, name, description):
        """Constructor."""
        self.__id = _id
        self._name = name
        self._description = description

    @property
    def _id(self):
        """Return the ID of the FrameNet corpus."""
        return self.__id

    @property
    def name(self):
        """Return the name of the FrameNet corpus."""
        return self._name

    @property
    def description(self):
        """Return the description of the FrameNet corpus."""
        return self._description



"""FrameNet FERelation class."""

__all__ = ['FERelation']


class FERelation():
    """FERelation."""

    def __init__(self, _id=None, sub_fe=None, sup_fe=None,
                 frame_relation=None):
        """Constructor."""
        self.__id = _id
        self._sub_fe = sub_fe
        self._sup_fe = sup_fe
        self._frame_relation = frame_relation

    @property
    def _id(self):
        return self.__id

    @property
    def sub_fe(self):
        """Return the sub FrameElement of the FERelation."""
        return self._sub_fe

    @property
    def sup_fe(self):
        """Return the sup FrameElement of the FERelation."""
        return self._sup_fe

    @property
    def frame_relation(self):
        """Return the FrameRelation of the FERelation."""
        return self._frame_relation



"""A class to store different format of valence pattern."""

__all__ = ['ValencePattern']


def _pt_key(valence_unit):
    return valence_unit.fe.itype if valence_unit.fe.itype else valence_unit.pt


def _sorted_valence_units(valence_units):
    return sorted(valence_units, key=lambda x: (x.fe.name, _pt_key(x), x.gf))


class ValencePattern():
    """ValencePattern."""

    def __init__(self, valence_units):
        """Constructor."""
        self._valence_units = valence_units

    @property
    def valence_units(self):
        """Return a list of valence units."""
        return self._valence_units

    @property
    def with_fe_name(self):
        """Return a stringified list of FE.PT.GF strings.

        The FrameElement fe follows: 'fe.name'
        """
        valence_pattern = ''
        if not self._valence_units:
            return valence_pattern
        for valence_unit in _sorted_valence_units(self._valence_units):
            valence_pattern = '{} {}'.format(valence_pattern,
                                             valence_unit.with_fe_name)
        return valence_pattern.strip()

    @property
    def with_fe_cnc(self):
        """Return a stringified list of FE.PT.GF strings.

        The FrameElement fe follows: 'fe.name#C' is the FrameElement is core
        or 'fe.name#NC' if the FrameElement is non-core
        """
        valence_pattern = ''
        if not self._valence_units:
            return valence_pattern
        for valence_unit in _sorted_valence_units(self._valence_units):
            valence_pattern = '{} {}'.format(valence_pattern,
                                             valence_unit.with_fe_cnc)
        return valence_pattern.strip()

    @property
    def with_fe_coretype(self):
        """Return a stringified list of FE.PT.GF strings.

        The FrameElement FE follows: 'fe.name#fe.coretype'
        """
        valence_pattern = ''
        if not self._valence_units:
            return valence_pattern
        for valence_unit in _sorted_valence_units(self._valence_units):
            valence_pattern = '{} {}'.format(valence_pattern,
                                             valence_unit.with_fe_coretype)
        return valence_pattern.strip()



"""XML processing error.

Error raised by the XML unmarshalling process
"""

__all__ = ['XMLProcessingError']


class XMLProcessingError(Exception):
    """A specific exception for invalid method."""

    def __init__(self, message):  # pylint:disable=W0235
        """Init function."""
        super().__init__(message)



"""Custom immutable objects."""

from pyfn.exceptions.method import InvalidMethodError
from pyfn.exceptions.parameter import InvalidParameterError

__all__ = ['FrozenDict', 'ImmutableConfig']


class FrozenDict(dict):
    """Immutable dict following the frozenset semantics."""

    def __setitem__(self, key, value):
        """Set method used for testing."""
        raise InvalidMethodError('Cannot assign value to a FrozenDict')


class ImmutableConfig(FrozenDict):
    """Immutable Configuration.

    Converts an input dict (of dict(s)) to a FrozenDict (of FrozenDict(s))
    """

    def __init__(self, config):
        """Constructor."""
        if not isinstance(config, dict):
            raise InvalidParameterError(
                'ImmutableConfig requires instance of dict as input parameter')
        super().__init__(self._freeze(config))

    def _get_frozen_value(self, input_value):
        if isinstance(input_value, dict):
            for key, value in input_value.items():
                if isinstance(value, dict):
                    input_value[key] = self._get_frozen_value(value)
            return FrozenDict(input_value)
        return input_value

    def _freeze(self, config):
        """Convert all dicts in config to FrozenDicts."""
        frozen_config = {}
        for key, value in config.items():
            frozen_config[key] = self._get_frozen_value(value)
        return FrozenDict(frozen_config)



"""FrameNet FrameElement class."""

__all__ = ['FrameElement']


class FrameElement():
    """FrameElement."""

    def __init__(self, _id, name, coretype=None, itype=None):
        """Constructor."""
        self.__id = _id
        self._name = name
        self._coretype = coretype
        self._itype = itype

    @property
    def _id(self):
        return self.__id

    @property
    def name(self):
        """Return FrameElement name."""
        return self._name

    @property
    def coretype(self):
        """Return FrameElement core type.

        Core Types include: 'Core', 'Core-Unexpressed', 'Peripheral' or
        'Extra-Thematic'. See FrameNet book for details.
        """
        return self._coretype

    @property
    def itype(self):
        """Return FrameElement instantiation type.

        itypes include 'INI', 'CNI' and 'DNI'. Seee FrameNet book for details.
        """
        return self._itype

    @coretype.setter
    def coretype(self, coretype):
        self._coretype = coretype

    @itype.setter
    def itype(self, itype):
        self._itype = itype



"""Unmarshall SemEval 2007 FrameNet XML files."""

import logging
import xml.etree.ElementTree as etree

import pyfn.marshalling.unmarshallers.framenet as fn_unmarshaller

__all__ = ['unmarshall_annosets']

logger = logging.getLogger(__name__)


def unmarshall_annosets(xml_filepath, fe_dict=None, flatten=False):
    """Unmarshall a SemEval 2007 FrameNet XML file from file path.

    Return a generator of AnnotationSet instances extracted from the
    XML file. A single list of AnnotationSet instances corresponds
    to all the AnnotationSets of a single Sentence.

    Args
    ----
        param1 xml_files_path: full path to a SemEval 2007 FrameNet XML file.

    """
    logger.info('Unmarshalling SemEval FrameNet XML file: {}'
                .format(xml_filepath))
    if fe_dict is None:
        fe_dict = {}
    # pylint: disable=R1702
    for documents_tag in etree.parse(xml_filepath).getroot().findall(
            'documents'):
        for document_tag in documents_tag.findall('document'):
            for paragraphs_tag in document_tag.findall('paragraphs'):
                for paragraph_tag in paragraphs_tag.findall('paragraph'):
                    for sentences_tag in paragraph_tag.findall('sentences'):
                        for sentence_tag in sentences_tag.findall('sentence'):
                            annosets = fn_unmarshaller.extract_fn_annosets(
                                sentence_tag, xml_schema_type='semeval',
                                fe_dict=fe_dict)
                            if annosets:
                                if not flatten:
                                    yield annosets
                                if flatten:
                                    for annoset in annosets:
                                        yield annoset



"""Docstring.

Details
"""

__all__ = ['InvalidMethodError']


class InvalidMethodError(Exception):
    """A specific exception for invalid method."""

    def __init__(self, message):  # pylint:disable=W0235
        """Init function."""
        super().__init__(message)



"""Config utilities.

Methods used to manipulate YAML-based configuration files.
"""

import logging
import yaml

from pyfn.utils.immutables import ImmutableConfig

__all__ = ['load']

logger = logging.getLogger(__name__)


def load(config_file):
    """Load an ImmutableConfig from a YAML configuration file."""
    logger.info('Loading config from file {}'.format(config_file))
    with open(config_file, 'r', encoding='utf-8') as config_stream:
        config = yaml.safe_load(config_stream)
        return ImmutableConfig(config)



"""Marshalling to SEMEVAL format.

The format matches the XML file expected by the perl evaluation script
of the SEMEVAL 2007 shared task on frame semantic structure extraction.
"""

import datetime
import logging
import pytz
import lxml.etree as etree

import pyfn.utils.filter as f_utils

from pyfn.exceptions.parameter import InvalidParameterError

__all__ = ['marshall_annosets']

logger = logging.getLogger(__name__)


# pylint: disable-msg=R0914
def _add_fe_labels(layers_tag, layer_id, annoset, label_id):
    fe_layer = etree.SubElement(layers_tag, 'layer')
    fe_layer.set('ID', str(layer_id))
    fe_layer.set('name', 'FE')
    fe_label_tags = etree.SubElement(fe_layer, 'labels')
    if 'FE' in annoset.labelstore.labels_by_layer_name:
        for fe_label in annoset.labelstore.labels_by_layer_name['FE']:
            if fe_label.start != -1 and fe_label.end != -1:
                fe_label_tag = etree.SubElement(fe_label_tags, 'label')
                fe_label_tag.set('ID', str(label_id))
                label_id += 1
                fe_label_tag.set('name', fe_label.name)
                fe_label_tag.set('start', str(fe_label.start))
                fe_label_tag.set('end', str(fe_label.end))
    return label_id


def _has_fe_labels(annoset):
    return 'FE' in annoset.labelstore.labels_by_layer_name


def _add_target_labels(layers_tag, layer_id, annoset, label_id):
    target_layer = etree.SubElement(layers_tag, 'layer')
    target_layer.set('ID', str(layer_id))
    target_layer.set('name', 'Target')
    target_labels = etree.SubElement(target_layer, 'labels')
    for target_index in annoset.target.indexes:
        target_label = etree.SubElement(target_labels, 'label')
        target_label.set('ID', str(label_id))
        target_label.set('name', 'Target')
        label_id += 1
        target_label.set('start', str(target_index[0]))
        target_label.set('end', str(target_index[1]))
    return label_id


def _get_annoset_tag(annosets_tag, annoset, annoset_id):
    annoset_tag = etree.SubElement(annosets_tag, 'annotationSet')
    annoset_tag.set('ID', str(annoset_id))
    annoset_tag.set('frameName', annoset.target.lexunit.frame.name)
    return annoset_tag


def _get_sentence_tag(annoset, sentences_tag, sent_id):
    sentence_tag = etree.SubElement(sentences_tag, 'sentence')
    sentence_tag.set('ID', str(sent_id))
    text = etree.SubElement(sentence_tag, 'text')
    text.text = annoset.sentence.text
    return sentence_tag


def _marshall_annosets(annosets, output_filepath, excluded_frames,
                       excluded_sentences, excluded_annosets):
    if not annosets:
        raise InvalidParameterError('No input annosets to marshall. Check '
                                    'input parameters and try again.')
    root = etree.Element('corpus')
    root.set('XMLCreated', datetime.datetime.now(
        pytz.utc).strftime('%a %b %d %H:%M:%S %Z %Y'))
    documents_tag = etree.SubElement(root, 'documents')
    document_tag = etree.SubElement(documents_tag, 'document')
    paragraphs_tag = etree.SubElement(document_tag, 'paragraphs')
    paragraph_tag = etree.SubElement(paragraphs_tag, 'paragraph')
    sentences_tag = etree.SubElement(paragraph_tag, 'sentences')
    sent_text = ''
    sent_id = 0  # to match the semval numbering of sentences
    annoset_id = 1
    layer_id = 1
    label_id = 1
    for annoset in f_utils.filter_and_sort_annosets(annosets, [],
                                                    excluded_frames,
                                                    excluded_sentences,
                                                    excluded_annosets):
        if annoset.sentence.text != sent_text:
            sentence = _get_sentence_tag(annoset, sentences_tag, sent_id)
            sent_id += 1
            sent_text = annoset.sentence.text
            annosets_tag = etree.SubElement(sentence, 'annotationSets')
        annoset_tag = _get_annoset_tag(annosets_tag, annoset, annoset_id)
        annoset_id += 1
        layers_tag = etree.SubElement(annoset_tag, 'layers')
        label_id = _add_target_labels(layers_tag, layer_id, annoset, label_id)
        layer_id += 1
        if _has_fe_labels(annoset):
            label_id = _add_fe_labels(layers_tag, layer_id, annoset, label_id)
            layer_id += 1
    tree = etree.ElementTree(root)
    tree.write(output_filepath, encoding='UTF-8', xml_declaration=True,
               pretty_print=True)


def marshall_annosets(annosets, output_filepath, excluded_frames,
                      excluded_sentences, excluded_annosets):
    """Marshall a list of pyfn.AnnotationSet objects to SEMEVAL XML.

    annosets: a list of annosets to marshall.
    output_filepath: the absolute path to the output .xml file
    excluded_frames: a list of frame #id to exclude from the output
    excluded_sentences: a list of sentence #id to exclude from the output
    excluded_annosets: a list of annotationset #id to exclude from the output
    """
    logger.info('Marshalling pyfn.AnnotationSet objects to SEMEVAL XML...')
    if not annosets:
        raise InvalidParameterError('Input pyfn.AnnotationSet list is empty')
    logger.info('Saving output to {}'.format(output_filepath))
    _marshall_annosets(annosets, output_filepath, excluded_frames,
                       excluded_sentences, excluded_annosets)



"""XML processing utils."""

import os

__all__ = ['extract', 'get_xml_filepaths']


def extract(directory):
    """Extract all filepaths ending with .xml from directory and return as list."""
    return [os.path.join(directory, filename) for filename in
            os.listdir(directory) if filename.endswith('.xml')
            and not filename.startswith('.')]


def get_xml_filepaths(splits_dirpath, subdir):
    """Return a list of fullpaths for XML files in splits_dirpath/subdir."""
    dirpath = os.path.join(splits_dirpath, subdir)
    if os.path.exists(dirpath):
        return extract(dirpath)
    return []



"""FrameNet FrameRelationType class."""

__all__ = ['FrameRelationType']


class FrameRelationType():
    """FrameRelationType."""

    def __init__(self, _id=None, name=None, sub_frame_name=None,
                 sup_frame_name=None):
        """Constructor."""
        self.__id = _id
        self._name = name
        self._sub_frame_name = sub_frame_name
        self._sup_frame_name = sup_frame_name

    @property
    def _id(self):
        return self.__id

    @property
    def name(self):
        """Return the name of the FrameRelationType."""
        return self._name

    @property
    def sub_frame_name(self):
        """Return the subFrameName of the FrameRelationType."""
        return self._sub_frame_name

    @property
    def sup_frame_name(self):
        """Return supFrameName of the FrameRelationType."""
        return self._sup_frame_name
